from dotenv import load_dotenv, find_dotenv

from ai_driver.pipelines import (
    pinecone_pipeline,
    local_llm_pipeline,
    local_download_pipeline,
)

load_dotenv(find_dotenv())

####
if __name__ == "__main__":
    pinecone_pipeline()
    local_llm_pipeline()
    local_download_pipeline()



import yaml
from langchain.vectorstores import FAISS
from langchain.text_splitter import RecursiveCharacterTextSplitter
from langchain.document_loaders import PyPDFLoader, DirectoryLoader
from langchain.embeddings import HuggingFaceEmbeddings
from loguru import logger
import box

# Import config vars
with open("config/config.yml", "r", encoding="utf8") as ymlfile:
    logger.info("Loading config")
    cfg = box.Box(yaml.safe_load(ymlfile))
    logger.info(cfg)


def run_db_build():
    # Build vector database
    loader = DirectoryLoader(cfg.DATA_PATH, glob="*.pdf", loader_cls=PyPDFLoader)
    documents = loader.load()
    text_splitter = RecursiveCharacterTextSplitter(
        chunk_size=cfg.CHUNK_SIZE, chunk_overlap=cfg.CHUNK_OVERLAP
    )
    texts = text_splitter.split_documents(documents)

    embeddings = HuggingFaceEmbeddings(
        model_name=cfg.CHAT_EMBED_MODEL,
        model_kwargs={"device": "cpu"},
    )

    vectorstore = FAISS.from_documents(texts, embeddings)
    vectorstore.save_local(cfg.DB_PATH)


if __name__ == "__main__":
    run_db_build()



from loguru import logger
from typing import Dict, Iterable
from langchain.vectorstores import FAISS
import time
from langchain.schema import Document
from langchain.embeddings import HuggingFaceInstructEmbeddings


def embed_FAISS_from_documents(
    documents: Iterable[Document],
    embedding_model_name: str,
    embedding_model_kwargs: Dict,
) -> FAISS:
    logger.info("Creating embeddings")
    start_time = time.time()
    instructor_embeddings = HuggingFaceInstructEmbeddings(
        model_name=embedding_model_name,
        model_kwargs=embedding_model_kwargs,
        cache_folder="",
    )
    end_time = time.time()
    elapsed_time = end_time - start_time
    logger.info(f"Embeddings created in {elapsed_time:.2f} seconds")

    logger.info("Creating vector store")
    start_time = time.time()
    db_instructEmbedd = FAISS.from_documents(documents, instructor_embeddings)
    end_time = time.time()
    elapsed_time = end_time - start_time
    logger.info(f"Vector store created in {elapsed_time:.2f} seconds")

    logger.info("Vector store created")
    return db_instructEmbedd



import os
from langchain.vectorstores import Pinecone
from langchain.embeddings.openai import OpenAIEmbeddings
import pinecone
from dataclasses import dataclass


@dataclass
class PineconeConfig:
    openai_key: str
    pinecone_key: str
    pinecone_env: str
    index_name: str


class PineconeSessionManager:
    """
    A class for managing Pinecone sessions and indexes.

    Attributes:
        embeddings (OpenAIEmbeddings): The embeddings object to use for indexing.
        index_name (str): The name of the Pinecone index to use.
        index (pinecone.GRPCIndex): The Pinecone index object.
        docsearch (Pinecone): The Pinecone search object.
    """

    def __init__(self, embeddings, config: PineconeConfig):
        """
        Initializes a new PineconeSessionManager instance.

        Args:
            embeddings (OpenAIEmbeddings): The embeddings object to use for indexing.
            index_name (str): The name of the Pinecone index to use.
        """
        self.embeddings = embeddings
        self.index_name = config.index_name
        # initialize pinecone
        pinecone.init(
            api_key=config.pinecone_key,  # find at app.pinecone.io
            environment=config.pinecone_env,  # next to api key in console
        )

        if self.index_name not in pinecone.list_indexes():
            raise ValueError(f"Index {self.index_name } not found")
        self.index = pinecone.GRPCIndex(self.index_name)
        self.docsearch: Pinecone = Pinecone.from_existing_index(
            index_name=self.index_name, embedding=embeddings
        )


def get_default_pinecone_session(config: PineconeConfig) -> PineconeSessionManager:
    """
    Returns a default PineconeSessionManager instance with OpenAI embeddings and a default index name.

    Returns:
        PineconeSessionManager: A new PineconeSessionManager instance.
    """
    embeddings = OpenAIEmbeddings(openai_api_key=config.openai_key)
    pc_session = PineconeSessionManager(embeddings, config)
    return pc_session



from dataclasses import dataclass
import box
import yaml


@dataclass
class InstructConfig:
    embed_model: str


def get_instruct_config():
    """Default Instruct Embedding Config"""
    with open("config/config.yml", "r", encoding="utf8") as ymlfile:
        cfg = box.Box(yaml.safe_load(ymlfile))
        config = InstructConfig(embed_model=cfg.INSTRUCT_EMBED_MODEL)
        return config



"""
This script downloads a model from Hugging Face and saves it to a specified output folder. It uses the Hugging Face API to get the download links for the model files and downloads them using multiple threads. The progress of the download is displayed using a progress bar.

The script takes in a list of models to download, the branch to download from, the number of threads to use for downloading, the output base folder, and the number of retries to attempt if the download fails.

The script defines the following functions:
- get_filename_from_url(url: str) -> str: Returns the filename from a given URL.
- get_output_path(output_folder: str, filename: str) -> str: Returns the output path for a given output folder and filename.
- get_http_response(url: str, headers: dict = {}, timeout: int = 10) -> requests.Response: Sends an HTTP GET request to the given URL with the specified headers and timeout and returns the response.
- write_content_to_file(output_path: str, data: bytes, mode: str = "ab") -> None: Writes the given data to a file at the specified output path with the specified mode.
- download_file(url: str, output_folder: str, progress_bar: Optional[Callable[[float, str], None]] = None) -> None: Downloads a file from a given URL and saves it to the specified output folder. If a progress bar function is provided, it will be called with the download progress and message.
- get_download_links(session: requests.Session, model: str, branch: str) -> List[str]: Gets the download links for the model files from the Hugging Face API for the given model and branch using the provided requests session.

The script can be run as a standalone script or imported as a module.
"""
from typing import Dict, Optional, Callable, Any, List
import base64
import json
import os
from pathlib import Path
from loguru import logger
import requests
import tqdm
from requests.adapters import HTTPAdapter
from tqdm.contrib.concurrent import thread_map
from requests import Session


def get_filename_from_url(url: str) -> str:
    """
    Extracts the filename from a given URL.

    Args:
        url (str): The URL to extract the filename from.

    Returns:
        str: The filename.
    """
    return os.path.basename(url)


def get_output_path(folder: str, filename: str) -> str:
    """
    Constructs the full output path for a file.

    Args:
        folder (str): The output folder to save the file in.
        filename (str): The name of the file.

    Returns:
        str: The full output path.
    """
    return os.path.join(folder, filename)


def get_http_response(
    url: str, headers: Dict[str, str] = None, timeout: int = 10
) -> requests.Response:
    """
    Makes an HTTP GET request to a URL and returns the response.

    Args:
        url (str): The URL to make the request to.
        headers (Dict[str, str], optional): The headers to include in the request. Defaults to {}.
        timeout (int, optional): The timeout for the request in seconds. Defaults to 10.

    Returns:
        requests.Response: The HTTP response.
    """
    if headers is None:
        headers = {}
    return requests.get(url, stream=True, headers=headers, timeout=timeout)


def write_content_to_file(output_path: str, data: bytes, mode: str = "ab") -> None:
    """
    Writes data to a file.

    Args:
        output_path (str): The path to the file to write to.
        data (bytes): The data to write.
        mode (str, optional): The mode to open the file in. Defaults to "ab".
    """
    with open(output_path, mode) as f:
        f.write(data)


def download_file(
    url: str, folder: str, progress_bar: Optional[Callable[..., Any]] = None
) -> None:
    """
    Downloads a file from a URL and saves it to an output folder, optionally updating a progress bar.

    Args:
        url (str): The URL of the file to download.
        output_folder (str): The folder to save the downloaded file in.
        progress_bar (Callable[..., Any], optional): A function to call to update a progress bar. Defaults to None.

    Raises:
        Exception: If there is an error downloading the file.
    """
    filename = get_filename_from_url(url)
    output_path = get_output_path(folder, filename)
    headers = {}
    block_size = 1024 * 1024  # 1MB
    logger.info(url)
    try:
        response = get_http_response(url, headers=headers)
        response.raise_for_status()
        logger.info(response)
        total_size = int(response.headers.get("content-length", 0))
        with tqdm.tqdm(
            total=total_size,
            unit="iB",
            unit_scale=True,
            bar_format="{l_bar}{bar}| {n_fmt:6}/{total_fmt:6} {rate_fmt:6}",
        ) as t:
            count = 0
            for data in response.iter_content(block_size):
                t.update(len(data))
                write_content_to_file(output_path, data)
                if total_size != 0 and progress_bar is not None:
                    count += len(data)
                    progress_bar(
                        float(count) / float(total_size),
                        f"Downloading {filename}",
                    )
    except Exception as err:
        logger.error(f"Error downloading file: {err}")
        raise


####


def construct_url(base: str, page: str, cursor: Optional[bytes] = None) -> str:
    """
    Constructs a URL given the base, page, and an optional cursor.

    Args:
        base (str): The base of the URL.
        page (str): The specific page after the base.
        cursor (Optional[bytes], optional): Optional cursor for pagination. Defaults to None.

    Returns:
        str: The full URL.
    """
    return f"{base}{page}" + (f"?cursor={cursor.decode()}" if cursor else "")


def make_request(session: Session, url: str) -> Dict:
    """
    Makes a GET request to a given URL and returns the JSON content.

    Args:
        session (Session): A requests Session object to make the request.
        url (str): The URL to make the request to.

    Returns:
        Dict: The JSON content of the response.
    """
    logger.info(f"Making request to: {url}")
    r = session.get(url, timeout=10)
    r.raise_for_status()
    return json.loads(r.content)


def extract_links(data_list: List[Dict], model: str, branch: str) -> List[str]:
    """
    Extracts download links from a list of file data.

    Args:
        data_list (List[Dict]): A list of dictionaries containing file data.
        model (str): The model name.
        branch (str): The branch name.

    Returns:
        List[str]: A list of download links.
    """
    links = []
    for file in data_list:
        fname = file["path"]
        if fname in ["1_Pooling", "2_Dense"]:
            logger.info(f"Subdir for sentence embed : {fname}")
            links.append(
                f"https://huggingface.co/{model}/resolve/{branch}/{fname}/config.json"
            )
            links.append(
                f"https://huggingface.co/{model}/resolve/{branch}/{fname}/pytorch_model.bin"
            )
            continue
        link = f"https://huggingface.co/{model}/resolve/{branch}/{fname}"
        logger.info(link)
        links.append(link)
    return links


def update_cursor(data_list: List[Dict]) -> bytes:
    """
    Updates the cursor for pagination.

    Args:
        data_list (List[Dict]): A list of dictionaries containing file data.

    Returns:
        bytes: The updated cursor.
    """
    cursor = (
        base64.b64encode(f'{{"file_name":"{data_list[-1]["path"]}"}}'.encode()) + b":50"
    )
    cursor = base64.b64encode(cursor)
    return cursor.replace(b"=", b"%3D")


def get_download_links(session: Session, model: str, branch: str) -> List[str]:
    """
    Gets download links for a given model and branch.

    Args:
        session (Session): A requests Session object to make the request.
        model (str): The model name.
        branch (str): The branch name.

    Returns:
        List[str]: A list of download links.
    """
    base = "https://huggingface.co"
    page = f"/api/models/{model}/tree/{branch}"
    cursor = b""
    links = []

    while True:
        url = construct_url(base, page, cursor)
        data_list = make_request(session, url)
        if len(data_list) == 0:
            break
        links.extend(extract_links(data_list, model, branch))
        cursor = update_cursor(data_list)

    return links


if __name__ == "__main__":
    # models = ["TheBloke/MPT-7B-Instruct-GGML", "TheBloke/Llama-2-7B-Chat-GGML"]
    # sentence_models = ["sentence-transformers/all-MiniLM-L6-v2", "hkunlp/instructor-xl"]
    test_model = ["google/flan-t5-base"]
    models = test_model
    branch = "main"
    threads = 6
    output_base = "artifacts"
    retries = 5

    for model in models:
        session = requests.Session()

        session.mount(
            "https://cdn-lfs.huggingface.co", HTTPAdapter(max_retries=retries)
        )
        session.mount("https://huggingface.co", HTTPAdapter(max_retries=retries))
        session.headers = {
            "authorization": f'Bearer {os.getenv("HUGGINGFACE_API_TOKEN")}'
        }

        output_folder = f"{'_'.join(model.split('/')[-2:])}"
        output_folder = Path(output_base) / output_folder

        links = get_download_links(session, model, branch)
        output_folder.mkdir(parents=True, exist_ok=True)
        logger.info(f"Downloading from url: https://huggingface.co/{model}")

        # Downloading the files
        print(f"Downloading the model to {output_folder}")
        thread_map(
            lambda url: download_file(url, output_folder),
            links,
            max_workers=threads,
            disable=True,
        )

        # with ThreadPoolExecutor(max_workers=threads) as executor:
        #     futures = []
        #     for url in links:
        #         futures.append(executor.submit(download_file, url, output_folder))
        #     for future in tqdm.tqdm(futures, desc="Downloading files", unit="file"):
        #         future.result()



import box
import yaml

from langchain import PromptTemplate
from langchain.chains import RetrievalQA
from langchain.embeddings import HuggingFaceEmbeddings
from langchain.vectorstores import FAISS
from ai_driver.local_llm.prompts import qa_template
from ai_driver.local_llm.ggml_llm import build_ggml_llm, get_default_ggml_config
from ai_driver.qa import QADBConfig


def set_qa_prompt():
    """
    Prompt template for QA retrieval for each vectorstore
    """
    prompt = PromptTemplate(
        template=qa_template, input_variables=["context", "question"]
    )
    return prompt


def build_retrieval_qa(llm, prompt, vectordb, config: QADBConfig):
    qa_chain = RetrievalQA.from_chain_type(
        llm=llm,
        chain_type="stuff",
        retriever=vectordb.as_retriever(search_kwargs={"k": config.vector_count}),
        return_source_documents=config.return_source,
        chain_type_kwargs={"prompt": prompt},
    )
    return qa_chain


def setup_local_qa_db(config: QADBConfig):
    embeddings = HuggingFaceEmbeddings(
        model_name=config.embed_model, model_kwargs={"device": "cpu"}
    )
    vectordb = FAISS.load_local(config.db_path, embeddings)

    ggml_config = get_default_ggml_config()
    llm = build_ggml_llm(ggml_config)
    qa_prompt = set_qa_prompt()
    dbqa = build_retrieval_qa(llm, qa_prompt, vectordb, config)

    return dbqa


def get_default_qa_config():
    # Import config vars
    with open("config/config.yml", "r", encoding="utf8") as ymlfile:
        cfg = box.Box(yaml.safe_load(ymlfile))
        config = QADBConfig(
            db_path=cfg.DB_PATH,
            embed_model=cfg.CHAT_EMBED_MODEL,
            return_source=cfg.RETURN_SOURCE_DOCUMENTS,
            vector_count=cfg.VECTOR_COUNT,
        )
        return config



from langchain.llms import CTransformers
import box
import yaml
from dataclasses import dataclass


@dataclass
class GGMLConfig:
    model_path: str
    model_type: str
    max_new_tokens: int
    temperature: float = 0.7


def build_ggml_llm(config: GGMLConfig) -> CTransformers:
    llm = CTransformers(
        model=config.model_path,
        model_type=config.model_type,
        config={
            "max_new_tokens": config.max_new_tokens,
            "temperature": config.temperature,
        },
    )
    return llm


def get_default_ggml_config():
    # Import config vars
    with open("config/config.yml", "r", encoding="utf8") as ymlfile:
        cfg = box.Box(yaml.safe_load(ymlfile))
        config = GGMLConfig(
            model_path=cfg.LLM_MODEL_BIN_PATH,
            model_type=cfg.MODEL_TYPE,
            max_new_tokens=cfg.MAX_NEW_TOKENS,
            temperature=cfg.TEMPERATURE,
        )
        return config



import os

LANGCHAIN_TRACING_V2 = os.environ.get("LANGCHAIN_TRACING_V2")
LANGCHAIN_PROJECT = os.environ.get("LANGCHAIN_PROJECT")
LANGCHAIN_ENDPOINT = os.environ.get("LANGCHAIN_ENDPOINT")
LANGCHAIN_API_KEY = os.environ.get("LANGCHAIN_API_KEY")

from langsmith import Client


def get_default_langsmith_client():
    """Default client for langchain logging"""
    return Client(api_url=LANGCHAIN_ENDPOINT, api_key=LANGCHAIN_API_KEY)



from typing import List, Iterable
from langchain.document_loaders import PyPDFLoader, DirectoryLoader
from loguru import logger
from langchain.text_splitter import RecursiveCharacterTextSplitter
from langchain.schema import Document


def load(dir_path: str, glob_pattern: str) -> Iterable[Document]:
    loader = DirectoryLoader(
        dir_path, glob=glob_pattern, loader_cls=PyPDFLoader
    )  # Note: If you're using PyPDFLoader then it will split by page for you already
    documents = loader.load()
    logger.info(f"Loaded {len(documents)} documents from {dir_path}")
    return documents


def split(documents: Iterable[Document], chunk_size, chunk_overlap) -> List[Document]:
    """
    Splits the specified list of PyPDFLoader instances into text chunks using a recursive character text splitter.

    Args:
        documents  (Iterable[Document]): The documents to split.
        chunk_size (int): The size of each text chunk.
        chunk_overlap (int): The overlap between adjacent text chunks.

    Returns:
        List[str]: A list of text chunks.
    """
    text_splitter = RecursiveCharacterTextSplitter(
        chunk_size=chunk_size, chunk_overlap=chunk_overlap
    )
    logger.info(type(documents))
    logger.info(
        f"Splitting {len(documents)} documents into chunks of size {chunk_size} with overlap {chunk_overlap}"
    )
    logger.info(documents)
    texts = text_splitter.split_documents(documents)
    logger.info(f"Split {len(documents)} documents into {len(texts)}")
    return texts


def get_default_local_download() -> List[Document]:
    """Default document list for local download"""
    dir_path = "data"
    glob_pattern = "./*.pdf"

    chunks = load(dir_path, glob_pattern)
    texts = split(chunks, chunk_size=1000, chunk_overlap=200)
    return texts



from ai_driver.local_loader import load, split
import pytest
import os
import shutil


@pytest.fixture
def chunks():
    return load(os.path.join(os.path.dirname(__file__), "test_data"), "*.pdf")


def test_should_return_empty_list_for_empty_directory():
    empty_dir = os.path.join(os.path.dirname(__file__), "test_data", "empty_dir")
    os.makedirs(empty_dir, exist_ok=True)
    assert load(empty_dir, "*.pdf") == []
    shutil.rmtree(empty_dir)


def test_should_return_empty_list_for_no_matching_files():
    assert load(os.path.join(os.path.dirname(__file__), "test_data"), "*.txt") == []


def test_should_return_chunks_from_matching_files():
    assert (
        len(load(os.path.join(os.path.dirname(__file__), "test_data"), "*.pdf")) == 115
    )


def test_should_split_chunks_by_size_and_overlap(chunks):
    chunk_size = 342
    chunk_overlap = 200

    texts = split(chunks, chunk_size, chunk_overlap)

    ### assert that chunks are not larger than expected
    for doc in texts:
        assert len(doc.page_content) <= chunk_size



import os
import timeit
from loguru import logger
from langchain.llms import OpenAI
from langchain.chains import RetrievalQA


from ai_driver.vector_storage.pinecone_manager import (
    get_default_pinecone_session,
    PineconeConfig,
)
from ai_driver.langsmith_config import get_default_langsmith_client
from ai_driver.vector_storage.faiss_manager import embed_FAISS_from_documents
from ai_driver.local_loader import get_default_local_download
from ai_driver.qa import query_documents
from ai_driver.instruct import get_instruct_config, InstructConfig
from ai_driver.local_llm.ggml_pipeline import setup_local_qa_db, get_default_qa_config


def pinecone_pipeline():
    """Example Pinecone Pipeline"""
    logger.info("Pinecone Pipeline")
    config = PineconeConfig(
        openai_key=os.getenv("OPENAI_API_KEY"),
        pinecone_key=os.getenv("PINECONE_API_KEY"),
        pinecone_env=os.getenv("PINECONE_API_ENV"),
        index_name=os.getenv("PINECONE_INDEX_NAME"),
    )
    logger.info(config)
    vector_store = get_default_pinecone_session(config).docsearch
    retriever = vector_store.as_retriever(search_kwargs={"k": 3})
    client = get_default_langsmith_client()
    qa_chain = RetrievalQA.from_chain_type(
        llm=OpenAI(client=client, temperature=0.2),
        chain_type="stuff",
        retriever=retriever,
        verbose=True,
        return_source_documents=True,
    )
    query_documents(qa_chain, "How do saving throws work?")


def local_llm_pipeline():
    """Example Local LLM Pipeline"""

    query = "How do saving throws work?"

    # Setup DBQA
    start = timeit.default_timer()
    config = get_default_qa_config()
    dbqa = setup_local_qa_db(config)
    response = dbqa({"query": query})
    end = timeit.default_timer()

    logger.info(f'\nAnswer: {response["result"]}')
    logger.info("=" * 50)

    # Process source documents
    source_docs = response["source_documents"]
    for i, doc in enumerate(source_docs):
        logger.info(f"\nSource Document {i+1}\n")
        logger.info(f"Source Text: {doc.page_content}")
        logger.info(f'Document Name: {doc.metadata["source"]}')
        logger.info(f'Page Number: {doc.metadata["page"]}\n')
        logger.info("=" * 60)

    logger.info(f"Time to retrieve response: {end - start}")


def local_download_pipeline(config=None):
    """Example local store pipeline"""
    logger.info("Local download pipeline: FAISS")
    texts = get_default_local_download()
    config: InstructConfig = get_instruct_config()

    embedding_model_name = config.embed_model
    embedding_model_kwargs = {"device": "cuda"}

    vector_store = embed_FAISS_from_documents(
        texts, embedding_model_name, embedding_model_kwargs
    )
    retriever = vector_store.as_retriever(search_kwargs={"k": 3})
    client = get_default_langsmith_client()
    qa_chain = RetrievalQA.from_chain_type(
        llm=OpenAI(client=client, temperature=0.2),
        chain_type="stuff",
        retriever=retriever,
        verbose=True,
        return_source_documents=True,
    )

    query_documents(qa_chain, "How do saving throws work?")



import os
import pytest
from ai_driver.scripts.download_model import (
    get_filename_from_url,
    get_output_path,
    write_content_to_file,
    construct_url,
    extract_links,
    update_cursor,
)


# Test get_filename_from_url
def test_get_filename_from_url_should_return_basename_of_url():
    url = "http://example.com/path/to/file.txt"
    assert get_filename_from_url(url) == "file.txt"


# Test get_output_path
def test_get_output_path_should_join_folder_and_filename():
    output_folder = "/path/to/folder"
    filename = "file.txt"
    assert get_output_path(output_folder, filename) == "/path/to/folder/file.txt"


# Test write_content_to_file
def test_write_content_to_file_should_write_content_to_file():
    output_path = "/tmp/test_file.txt"  # Use a temporary file for the test
    data = b"test content"
    write_content_to_file(output_path, data)

    with open(output_path, "rb") as f:
        assert f.read() == data

    os.remove(output_path)  # Clean up the temporary file after the test


# Test construct_url
def test_construct_url_should_return_url_with_cursor_when_cursor_is_provided():
    base = "https://huggingface.co"
    page = "/api/models/model/tree/branch"
    cursor = b"cursor"
    expected_url = "https://huggingface.co/api/models/model/tree/branch?cursor=cursor"
    assert construct_url(base, page, cursor) == expected_url


# Test extract_links
def test_extract_links_should_extract_links_for_pooling_and_dense_directories_and_other_files():
    data_list = [
        {"path": "1_Pooling"},
        {"path": "2_Dense"},
        {"path": "other"},
    ]
    model = "model"
    branch = "branch"
    expected_links = [
        "https://huggingface.co/model/resolve/branch/1_Pooling/config.json",
        "https://huggingface.co/model/resolve/branch/1_Pooling/pytorch_model.bin",
        "https://huggingface.co/model/resolve/branch/2_Dense/config.json",
        "https://huggingface.co/model/resolve/branch/2_Dense/pytorch_model.bin",
        "https://huggingface.co/model/resolve/branch/other",
    ]
    assert extract_links(data_list, model, branch) == expected_links


# Test update_cursor
def test_update_cursor_should_return_base64_encoded_filename_and_cursor_string():
    data_list = [{"path": "file1"}, {"path": "file2"}]
    # The expected_cursor is hardcoded after manually computing the expected result
    expected_cursor = b"ZXlKbWFXeGxYMjVoYldVaU9pSm1hV3hsTWlKOTo1MA%3D%3D"
    assert update_cursor(data_list) == expected_cursor



qa_template = """Use the following pieces of information to answer the user's question.
If you don't know the answer, just say that you don't know, don't try to make up an answer.

Context: {context}
Question: {question}

Only return the helpful answer below and nothing else.
Helpful answer:
"""



import textwrap
from ai_driver.qa import wrap_text_preserve_newlines


def test_wrap_text_preserve_newlines():
    text = "This is a long line of text that should be wrapped to fit within the specified width. It also contains\na newline character that should be preserved."
    expected_output = "This is a long line of text that should be wrapped\nto fit within the specified width. It also\ncontains\na newline character that should be preserved."
    assert wrap_text_preserve_newlines(text, width=50) == expected_output

    text = "This is another long line of text that should be wrapped to fit within the specified width. It also contains\nmultiple newline characters that should be preserved.\n\nThis is a new paragraph that should be wrapped separately."
    expected_output = "This is another long line of text that\nshould be wrapped to fit within the\nspecified width. It also contains\nmultiple newline characters that should\nbe preserved.\n\nThis is a new paragraph that should be\nwrapped separately."
    assert wrap_text_preserve_newlines(text, width=40) == expected_output



import textwrap
from dataclasses import dataclass

from loguru import logger


@dataclass
class QADBConfig:
    db_path: str
    embed_model: str
    return_source: str
    vector_count: str


def wrap_text_preserve_newlines(text, width=110):
    lines = text.split("\n")
    wrapped_lines = [textwrap.fill(line, width=width) for line in lines]
    wrapped_text = "\n".join(wrapped_lines)
    return wrapped_text


def process_llm_response(llm_response):
    logger.info(wrap_text_preserve_newlines(llm_response["result"]))
    logger.info("\nSources:")
    for source in llm_response["source_documents"]:
        logger.info(source.metadata)


def query_documents(qa_chain, query):
    llm_response = qa_chain(query)
    process_llm_response(llm_response)
